/*
 * UNC0xx host-side software. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc0xx.h
 * Headers of public functions for communicating with the device
 *
 * Written by Andrey V. Stolyarov
 *
 */


#ifndef UNC0XX_H_SENTRY
#define UNC0XX_H_SENTRY

#ifndef UNC_INPUTS_COUNT
#define UNC_INPUTS_COUNT 8
#endif

#include <stdint.h>


#ifdef __cplusplus
extern "C" {
#endif

    /* The generic device handler, which is independent of the
     * communication method.
     * This structure is opaque, that is, you shouldn't want to use
     * its fields directly.  The fields can be changed in future
     * versions of the library, without a notice.
     */
struct unc0xx_devhdl;


/* Now the USB-related stuff goes =================================== */

    /* This function must be called before (and if) you use devices
     * connected directly via USB */
    /* Actually, it initializes global data of the libusb library,
     * so in case your application uses libusb for communicating
     * devices other than UNC0xx, you should only initialize it once,
     * but make sure the initialization happens, one way or another.
     * If you only use USB to communicate UNC0xx within your program,
     * that is, you don't call libusb directly, then just ignore all
     * this stuff and simply call this function.
     */
void unc0xx_usb_init();

    /* Call this function to rescan the bus for connected devices */
void unc0xx_usb_rescan();
    /* This structure is used to form a simple unidirection list to
       store information about all the UNC0xx devices found on your bus.
     */
struct unc0xx_scan_item {
    int sernum[3];
    int version;                   /* 0 for devs with deprecated protocols */
    struct unc0xx_devhdl *dev_hdl; /* NULL for deprecated/incompatible devs */
    struct unc0xx_scan_item *next;
};

    /* the list MUST be destroyed later with unc0xx_destroy_list */
struct unc0xx_scan_item *unc0xx_scan(int unc0xx_id);
    /* all dev_hdl's are closed; to avoid this, assign NULL to them */
void unc0xx_destroy_list(struct unc0xx_scan_item *lst);

extern int unc0xx_usb_timeout;

/* End of the USB-related stuff ===================================== */


/* ================================================================== */


/* Now the RS485-specific stuff ===================================== */

    /* Creates a handler for communication over rs485 */
    /* In case of troubles, NULL is returned.  Please note that
     * 'troubles' only mean that the port couldn't be opened; if
     * the port was opened successfully, the function doesn't check
     * whether the device is actually present.  Try querying the
     * device to check whether it is there.
     *
     * WARNING: if you open multiple devices, the handles will
     * share the same file descriptor, which is achieved by naive
     * comparision of port file names.  This approach isn't very
     * correct as the file (theoretically) can be renamed, removed
     * or whatever.  In the real life, however, such things never happen.
     *
     * Please note also that baud rate on an already open port is never
     * changed.  The parameter is ignored if the port is already open.
     * If unsure, specify 0 for baud, which means the default 19200.
     */
struct unc0xx_devhdl* unc0xx_rs485_open(const char *pname, int baud, int id);

struct unc0xx_devhdl* unc0xx_rs485_probe(const char *pname, int baud,
                                         int id, int *vers);


/* End of the RS485-specific stuff ================================== */



    /* The close function is common for all possible communication methods */
void unc0xx_close(struct unc0xx_devhdl *dev_hdl);

    /* Works ONLY after the first successfull configuration request  */
    /* doesn't request anything, just returns internally stored info */
int unc0xx_get_version(struct unc0xx_devhdl *dev_hdl);

struct unc0xx_global_configuration_str;
int unc0xx_get_configuration(struct unc0xx_devhdl *dev_hdl,
                             struct unc0xx_global_configuration_str *data);

struct unc0xx_global_state_str;
int unc0xx_get_fullstate(struct unc0xx_devhdl *dev_hdl,
                         struct unc0xx_global_state_str *data);

struct unc0xx_extra_state_str_common;
int unc0xx_get_extrastate(struct unc0xx_devhdl *dev_hdl,
                          struct unc0xx_extra_state_str_common *data);


int unc0xx_save_to_eeprom(struct unc0xx_devhdl *dev_hdl);
int unc0xx_restore_from_eeprom(struct unc0xx_devhdl *dev_hdl);
int unc0xx_restore_factory(struct unc0xx_devhdl *dev_hdl);

int unc0xx_clear_msg(struct unc0xx_devhdl *dev_hdl);
int unc0xx_reset_1wire(struct unc0xx_devhdl *dev_hdl);

int unc0xx_set_new_id(struct unc0xx_devhdl *dev_hdl, unsigned int id);
int unc0xx_set_configuration(struct unc0xx_devhdl *dev_hdl, unsigned int conf);
    /* use with EXTREME CAUTION!!! better don't use at all */
int unc0xx_set_hardware_conf(struct unc0xx_devhdl *dev_hdl, unsigned int conf);

int unc0xx_perform_action(struct unc0xx_devhdl *dev_hdl, int action);
int unc0xx_delayed_action(struct unc0xx_devhdl *dev_hdl,
                          int action, int delay, int location);
int unc0xx_set_reactions(struct unc0xx_devhdl *dev_hdl,
                         int line, unsigned long reaction);
int unc0xx_get_counter_range(struct unc0xx_devhdl *dev_hdl,
                         int *min, int *max);
int unc0xx_zero_query_counters(struct unc0xx_devhdl *dev_hdl,
                               int counters[], int ncount);

    /* buf must point to a buffer of 128 bytes to support versions
       prior to 4xxxxxx; modern versions, however, use only 64 of them */
int unc0xx_dump_eeprom(struct unc0xx_devhdl *dev_hdl, void* buf);


    /* Prepares to read the "interrupt input" from the device.
     * With USB, this might require to tell the system driver that the
     * stream is no longer to be handled by the system.  With RS485, this
     * is not supported at all; with Windows, it simply doesn't work.
     */
int unc0xx_takeover_stream(struct unc0xx_devhdl *dev_hdl);

    /* Tries to read the "interrupt input" from the device.
     * The buffer must be exactly 8 bytes long!
     * Function returns the actual number of received bytes,
     * or -1 on error.  Not supported for RS485; doesn't work
     * with Windows.
     */
int unc0xx_read_stream(struct unc0xx_devhdl *dev_hdl, char buf[8]);

    /* Overwrites the given Exx register */
int unc0xx_write_e_reg(struct unc0xx_devhdl *dev_hdl, int regn, char data[8]);

const char *unc0xx_get_errormsg(struct unc0xx_devhdl *h);

#ifdef __cplusplus
}
#endif

#endif
