/*
 * UNC0xx host-side software. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc0xx_rs485.c
 * rs485-specific stuff for communicating with the device
 *
 * Written by Andrey V. Stolyarov
 *
 */

#ifdef _WIN32

#include <windows.h>

#else

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#endif

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>



#include "unc_data.h"

#include "unc0xx.h"
#include "unc0xx_hdl.h"


enum {
    max_response_len = 256,
    timeout_msec = 2000,
    pause_after_input_ms = 150
};

/* ------------------- the handlers ------------------ */

struct com_port_hdl {
    const char *name;
#ifdef _WIN32
    HANDLE win_handle;
#else
    int fd;
    struct timeval last_input;
#endif
    int refcount;
    struct com_port_hdl *next;
};

static struct com_port_hdl *first_porthdl = NULL;

static int com_port_speed(int bdr)
{
#ifdef _WIN32
    switch(bdr) {
        case   4800:  return   CBR_4800;
        case   9600:  return   CBR_9600;
        case  19200:  return  CBR_19200;
        case  38400:  return  CBR_38400;
        case  57600:  return  CBR_57600;
        /* case 115200:  return CBR_115200; */

        default:      return  CBR_19200;
    }
#else
    switch(bdr) {
        case   4800:  return   B4800;
        case   9600:  return   B9600;
        case  19200:  return  B19200;
        case  38400:  return  B38400;
        case  57600:  return  B57600;
        /* case 115200:  return B115200; */

        default:      return  B19200;
    }
#endif
}

static struct com_port_hdl *com_port_connect(const char *name, int baud)
{
    struct com_port_hdl *tmp;
#ifdef _WIN32
    HANDLE win_handle;
    DCB port_params = { 0 };
    COMMTIMEOUTS timeout = { 0 };
#else
    int fd;
    struct termios tios;
#endif

    if(!baud)
        baud = 19200;

    /* check whether it is already there */
    for(tmp = first_porthdl; tmp; tmp = tmp->next)
        if(0 == strcmp(name, tmp->name)) {
            tmp->refcount++;
            return tmp;  /* already connected */
        }

    /* okay, we've got to make another connection */

#ifdef _WIN32

    win_handle = CreateFile(name, GENERIC_READ | GENERIC_WRITE, 0,
                 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

    port_params.DCBlength = sizeof(port_params);
    GetCommState(win_handle, &port_params);
    port_params.BaudRate = com_port_speed(baud);
    port_params.ByteSize = 8;
    port_params.StopBits = ONESTOPBIT;
    port_params.Parity = NOPARITY;
    SetCommState(win_handle, &port_params);

    timeout.ReadIntervalTimeout = 50;
    timeout.ReadTotalTimeoutConstant = timeout_msec;
    timeout.ReadTotalTimeoutMultiplier = 50;
    timeout.WriteTotalTimeoutConstant = 50;
    timeout.WriteTotalTimeoutMultiplier = 10;
    SetCommTimeouts(win_handle, &timeout);

#else

    fd = open(name, O_RDWR | O_NOCTTY);
    if(fd == -1)
        return NULL;
    if(-1 == tcgetattr(fd, &tios)) {
        close(fd);
        return NULL;
    }
    cfsetispeed(&tios, com_port_speed(baud));
    cfsetospeed(&tios, com_port_speed(baud));
    tios.c_iflag &=
        ~(IGNBRK | BRKINT | ICRNL | INLCR | PARMRK | INPCK | ISTRIP | IXON);
    tios.c_oflag = 0;
    tios.c_lflag &= ~(ECHO | ECHONL | ICANON | IEXTEN | ISIG);
    tios.c_cflag &= ~(CSIZE | PARENB);
    tios.c_cflag |= CS8;
    tios.c_cc[VMIN]  = 1;
    tios.c_cc[VTIME] = 0;
    tcsetattr(fd, TCSANOW, &tios);
    tcflush(fd, TCIOFLUSH);

#endif

    /* now place it into the list */
    tmp = malloc(sizeof(*tmp));
    tmp->name = strdup(name);
#ifdef _WIN32
    tmp->win_handle = win_handle;
#else
    tmp->fd = fd;
    tmp->last_input.tv_sec = 0;
    tmp->last_input.tv_usec = 0;
#endif
    tmp->refcount = 1;
    tmp->next = first_porthdl;
    first_porthdl = tmp;

    return tmp;
}

static void com_port_release(struct com_port_hdl *h)
{
    struct com_port_hdl **tmp;

    if(--(h->refcount) >= 1)
        return;

#ifdef _WIN32
    CloseHandle(h->win_handle);
#else
    close(h->fd);
#endif

    free((void*)h->name);

    tmp = &first_porthdl;
    while(*tmp && *tmp != h)
        tmp = &((*tmp)->next);
    if(*tmp)
        *tmp = h->next;

    free(h);
}

enum {
    com_port_read_timeout = -2,
    com_port_read_error = -1,
    com_port_read_eof = 0
};

static int com_port_read(struct com_port_hdl *h, void *buf, int size)
{
#ifdef _WIN32
    /* no need for select() in windows, the timeouts handled by the comm */
    DWORD bytes_read = 0;
    if(!ReadFile(h->win_handle, buf, size, &bytes_read, NULL))
        return com_port_read_error;
    if(!bytes_read)
        return com_port_read_eof;
    return bytes_read;

#else

    struct timeval tv;
    fd_set read_set;
    int rc;
    tv.tv_sec = timeout_msec / 1000;
    tv.tv_usec = (timeout_msec % 1000) * 1000;
    FD_ZERO(&read_set);
    FD_SET(h->fd, &read_set);
    rc = select(h->fd+1, &read_set, NULL, NULL, &tv);
    if(rc == 0)
        return com_port_read_timeout;
    rc = read(h->fd, buf, size);
    if(rc < 0)
        return com_port_read_error;
    if(rc == 0)
        return com_port_read_eof;
    gettimeofday(&h->last_input, NULL);
    return rc;
#endif
}

static void com_port_wait_if_needed(struct com_port_hdl *h)
{
    /* we need this function because, after the dev->host transmission is
       done, the device remains in TX mode some time, smth. like 1/20
       a second
     */

#ifdef _WIN32
    Sleep(pause_after_input_ms);
    /* this must change! we need to measure the time since the last
       read from the port, and wait only as long as we need it to
       maintain the pause between transactions
     */
#else
    struct timeval tv;
    long long delta;

    if(!h->last_input.tv_sec)
        return;               /* no need for pause */
    gettimeofday(&tv, NULL);
    delta =
        (long long)(tv.tv_sec - h->last_input.tv_sec) * 1000000LL +
        (long long)(tv.tv_usec - h->last_input.tv_usec);
    if(delta < pause_after_input_ms * 1000) {
        usleep(pause_after_input_ms * 1000 - delta);
    }
    h->last_input.tv_sec = 0;
#endif
}

static int com_port_write(struct com_port_hdl *h, const void *buf, int size)
{
#ifdef _WIN32
    DWORD bytes_written = 0;
    if(!WriteFile(h->win_handle, buf, size, &bytes_written, NULL))
        return -1;
    return bytes_written;

#else

    int wc;
    /*set_dtr(fd, 0);*/
    wc = write(h->fd, buf, size);
    /*tcdrain(h->fd);*/
    /*usleep((p-buf+1)*8 * 1000000 / 9600);*/
    /*set_dtr(fd, 1);*/
    /*printf("@\n");*/
    return wc;
#endif
}

struct unc0xx_devhdl_rs485 {
    struct unc0xx_devhdl parent;
    struct com_port_hdl *port;
    int id;
    struct unc0xx_devhdl_rs485 *next;
};

static struct unc0xx_devhdl_rs485 *first_handler = NULL;

static int rs485_method_get(struct unc0xx_devhdl *, int, void *, int);
static int rs485_method_send(struct unc0xx_devhdl *, const void *, int);
static int rs485_method_close(struct unc0xx_devhdl *);
static int rs485_method_takeover_stream(struct unc0xx_devhdl *);
static int rs485_method_read_stream(struct unc0xx_devhdl *, char *);

static struct unc0xx_devhdl_methods the_rs485_methods = {
    rs485_method_get,
    rs485_method_send,
    rs485_method_close,
    rs485_method_takeover_stream,
    rs485_method_read_stream
};

static struct unc0xx_devhdl_rs485 *
create_handler(struct com_port_hdl *port, int id)
{
    struct unc0xx_devhdl_rs485 *tmp = malloc(sizeof(*tmp));
    tmp->parent.vtbl = &the_rs485_methods;
    tmp->parent.version = 0;
    tmp->parent.message = NULL;
    tmp->port = port;
    tmp->id = id;
    tmp->next = first_handler;
    first_handler = tmp;
    return tmp;
}

static void destroy_handler(struct unc0xx_devhdl_rs485 *handler)
{
    if(handler->port) {
        com_port_release(handler->port);
        handler->port = NULL;
    }
    struct unc0xx_devhdl_rs485 **tmp = &first_handler;
    while(*tmp && *tmp != handler)
        tmp = &((*tmp)->next);
    if(*tmp)
        *tmp = handler->next;
    if(handler->parent.message && handler->parent.message_owned)
        free((char *)(handler->parent.message));
    free(handler);
}


/* the rs485 implementation ------------------------------------- */

static char hexdigit(int n) {
    if(n<=9) return n + '0';
    if(n<=15) return n - 10 + 'A';
    else return '?'; /* should never happen */
}

static int hexdig2num(char c)
{
    if(c>='0' && c<='9')
        return c - '0';
    if(c>='a' && c<='f')
        return c - 'a' + 10;
    if(c>='A' && c<='F')
        return c - 'A' + 10;
    return -1;
}

static void send_command(struct unc0xx_devhdl_rs485 *hdl,
                         const unsigned char *c, int clen)
{
    static unsigned char buf[64];
    unsigned char *p = buf;
    p += sprintf((char*)buf, ":%04X", hdl->id);
    while(clen) {
        *p++ = hexdigit((*c >> 4) & 0x0f);
        *p++ = hexdigit(*c & 0x0f);
        *p++ = hexdigit((*c ^ (*c >> 4)) & 0x0f);
        c++;
        clen--;
    }
    *p++ = '\r';
    *p++ = '\n';
    *p = 0;
    com_port_wait_if_needed(hdl->port);
    com_port_write(hdl->port, buf, p - buf);
}

static void send_query_command(struct unc0xx_devhdl_rs485 *hdl,
                               unsigned char quid)
{
    send_command(hdl, &quid, 1);
}

static int get_response(struct unc0xx_devhdl_rs485 *hdl,
                        unsigned char buf[max_response_len])
{
    unsigned char rdbuf[max_response_len / 2];
    int result_len = 0;
    int msgpos = -1;
    int remote_id = 0;
    int rc, i, dig;
    for(;;) {
        rc = com_port_read(hdl->port, rdbuf, sizeof(rdbuf));
        switch(rc) {
        case com_port_read_timeout:
            unc0xx_set_errormsg(&hdl->parent,
                "com port reading timed out", unc0xx_msg_const);
            return -1;
        case com_port_read_error:
            unc0xx_set_errormsg(&hdl->parent,
                 strerror(errno), unc0xx_msg_copy);
            return -1;
        case com_port_read_eof:
            unc0xx_set_errormsg(&hdl->parent,
                "unexpected EOF on com port", unc0xx_msg_const);
            return -1;
        default:
            ;
        }
        for(i = 0; i < rc; i++) {
            /*printf("%c", rdbuf[i]);*/
            if(rdbuf[i] == '\n') {
                if(msgpos < 5) {
                    unc0xx_set_errormsg(&hdl->parent,
                        "response message too short", unc0xx_msg_const);
                    return -1;
                }
                return result_len;
            }
            if(isspace(rdbuf[i]))
                continue;
            if(msgpos == -1) {
                if(rdbuf[i] == ':')
                    msgpos = 0;
                continue;
            }
            dig = hexdig2num(rdbuf[i]);
            if(msgpos < 4) {
                remote_id = (remote_id << 4) | dig;
                if(++msgpos == 4) {
                    if(remote_id != hdl->id) {
                        char *err = malloc(32);
                        sprintf(err, "id mismatch %04x vs. %04x",
                                      remote_id, hdl->id);
                        unc0xx_set_errormsg(&hdl->parent, err, unc0xx_msg_own);
                        return -1;
                    }
                }
                continue;
            }
            switch((msgpos - 4) % 3) {
                case 0:
                    buf[result_len] = dig << 4;
                    break;
                case 1:
                    buf[result_len] |= dig;
                    break;
                case 2:
                    if(((buf[result_len]>>4)^(buf[result_len]&0x0f)) != dig) {
                        char *err = malloc(32);
                        sprintf(err, "CRC mismatch at %d(%d)",
                                msgpos, result_len);
                        unc0xx_set_errormsg(&hdl->parent, err, unc0xx_msg_own);
                        return -1;
                    }
                    result_len++;
            }
            msgpos++;
        }
    }
}

/* ------------------- the methods ------------------ */


static int rs485_method_close(struct unc0xx_devhdl *h)
{
    struct unc0xx_devhdl_rs485 *hdl = (struct unc0xx_devhdl_rs485 *)h;
    com_port_release(hdl->port);
    hdl->port = NULL;
    destroy_handler(hdl);
    return 0;
}


static
int rs485_method_get(struct unc0xx_devhdl *h, int query_id, void *buf, int len)
{
    unsigned char mbuf[max_response_len];
    int rc, i;
    struct unc0xx_devhdl_rs485 *hdl = (struct unc0xx_devhdl_rs485 *)h;
    send_query_command(hdl, query_id);
    rc = get_response(hdl, mbuf);
    if(rc <= 0)
        return rc;
    if(len < rc)
        rc = len;
    for(i = 0; i < rc; i++)
        ((char*)buf)[i] = mbuf[i];
    return rc;
}


static
int rs485_method_send(struct unc0xx_devhdl *h, const void *buf, int len)
{
    int rc;
    unsigned char resp[max_response_len];
    struct unc0xx_devhdl_rs485 *hdl = (struct unc0xx_devhdl_rs485 *)h;
    send_command(hdl, buf, len);
    rc = get_response(hdl, resp);
    return (rc == 1 && resp[0] == 0) ? len : -1;
}

static int rs485_method_takeover_stream(struct unc0xx_devhdl *h)
{
    unc0xx_set_errormsg(h,
        "Input streams are not implemented for rs485", unc0xx_msg_const);
    return -1;
}

static int rs485_method_read_stream(struct unc0xx_devhdl *h, char *buf)
{
    unc0xx_set_errormsg(h,
        "Input streams are not implemented for rs485", unc0xx_msg_const);
    return -1;
}


/* The Public Function ------------------------------ */

struct unc0xx_devhdl* unc0xx_rs485_open(const char *pname, int baud, int id)
{
    struct com_port_hdl *port = com_port_connect(pname, baud);
    if(!port)
        return NULL;
    return (struct unc0xx_devhdl*)create_handler(port, id);
}

struct unc0xx_devhdl*
unc0xx_rs485_probe(const char *pname, int baud, int id, int *version)
{
    int r;
    struct unc0xx_global_configuration_str s1;
    struct unc0xx_devhdl *hdl = unc0xx_rs485_open(pname, baud, id);
    if(!hdl)
        return NULL;

    r = unc0xx_get_configuration(hdl, &s1);
    *version = (r == -1) ? 0 : s1.version;

    return hdl;
}
