/*
 * UNC0xx host-side software. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc0xx_usb.c
 * USB-specific stuff for communicating with the device
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdlib.h>   /* for malloc/free */
#include <string.h>   /* for strcmp */

#ifdef _WIN32
#define USE_HIDAPI
#endif

#ifndef USE_HIDAPI
#include <usb.h>
#else
#include <windows.h>
#include <setupapi.h>
#include "hidapi.h"
#endif

#include "unc_id.h"
#include "unc_data.h"

#include "unc0xx.h"
#include "unc0xx_hdl.h"


/* ------------------- the handlers ------------------ */

struct unc0xx_devhdl_usb {
    struct unc0xx_devhdl parent;
    void *dev_hdl;
    struct unc0xx_devhdl_usb *next;
};

static struct unc0xx_devhdl_usb *first_handler = NULL;

static int usb_method_get(struct unc0xx_devhdl *, int, void *, int);
static int usb_method_send(struct unc0xx_devhdl *, const void *, int);
static int usb_method_close(struct unc0xx_devhdl *);
static int usb_method_takeover_stream(struct unc0xx_devhdl *);
static int usb_method_read_stream(struct unc0xx_devhdl *, char *);

static struct unc0xx_devhdl_methods the_usb_methods = {
    usb_method_get,
    usb_method_send,
    usb_method_close,
    usb_method_takeover_stream,
    usb_method_read_stream
};

static struct unc0xx_devhdl_usb *create_handler()
{
    struct unc0xx_devhdl_usb *tmp = malloc(sizeof(*tmp));
    tmp->parent.vtbl = &the_usb_methods;
    tmp->parent.version = 0;
    tmp->parent.message = NULL;
    tmp->dev_hdl = NULL;
    tmp->next = first_handler;
    first_handler = tmp;
    return tmp;
}

static void destroy_handler(struct unc0xx_devhdl_usb *handler)
{
    struct unc0xx_devhdl_usb **tmp = &first_handler;
    while(*tmp && *tmp != handler)
        tmp = &((*tmp)->next);
    if(*tmp)
        *tmp = handler->next;
    if(handler->parent.message && handler->parent.message_owned)
        free((char *)(handler->parent.message));
    free(handler);
}



/* ------------------- usb comm. implementation ------------------ */

static const char *compatible_vendor_names[] = {
    UNC_VENDOR_NAME,
    "UNIControllers http://www.unicontrollers.com",
    NULL
};

static const char *compatible_device_names[] = {
    UNC_DEVICE_NAME,
    UNC_DEVICE_NAME_0XX_LEGACY,
    UNC_DEVICE_NAME_001_LEGACY,
    NULL
};

static const char *deprecated_device_names[] = {
    UNC_DEVICE_NAME_001_DEPRECATED,
    NULL
};

#ifndef USE_HIDAPI  /* this stuff is not needed for hidapi */
/* shamelessly stolen from Christian Starkjohann's hiddata.c file */
#define USBRQ_HID_GET_REPORT    0x01
#define USBRQ_HID_SET_REPORT    0x09
#define USB_HID_REPORT_TYPE_FEATURE 3
#endif

/* public (global) variable */
/* please note it is not used with hidapi */
int unc0xx_usb_timeout = 5000;

void unc0xx_usb_init()
{
#ifdef USE_HIDAPI
    /* no initialization required for hidapi */
#else
    usb_init();
    usb_find_busses();
    usb_find_devices();
#endif
}

void unc0xx_usb_rescan()
{
#ifdef USE_HIDAPI
    /* no global variables are used for hidapi; rescan is a no-op */
#else
    usb_find_devices();
#endif
}

static int hexdig2num(char c)
{
    if(c >= '0' && c <= '9') return c - '0';    
    if(c >= 'A' && c <= 'F') return c - 'A' + 10;    
    if(c >= 'a' && c <= 'f') return c - 'a' + 10;    
    return -1;
}

static int extract_sernum(const char *str, int sernum[3])
{
    int p = 0;
    int n = 0;
    sernum[0] = sernum[1] = sernum[2] = -1;
    for(; *str && p<3; str++) {
        if(*str == ':') {
            sernum[p] = n;
            p++;
            n = 0;
        } else {
            int d = hexdig2num(*str);
            if(d == -1) return 0;
            n = n*16 + d;
        }
    }
    if(!*str) {
        sernum[p] = n;
        p++;
    }
    return p==3;
}


#ifdef USE_HIDAPI
/* hidapi uses wchar_t arrays but this is senseless for 
   USB strings as they SHOULD contain ASCII chars only
   unless the vendor goes totally mad.  Hell if there
   is a point to make your device unusable for all the
   world outside of your non-latin country; be wise,
   use English.  What we definitely can say is that
   Unicontrollers' devices will ALWAYS return ASCII
   strings without non-ascii chars.
   Anyway, what we need is a convertor.
*/
void simple_from_wchar(const wchar_t *array, char buf[80])
{
   int i;
   if(!array) {
      buf[0]=0;
      return;
   }
   for(i=0; i<79 && array[i]; i++)
       buf[i] = array[i]>127 ? '?' : array[i];
   buf[i] = 0;
}

#endif

static int check_vendor_name(const char *str)
{
    const char * const *p;
    for(p = compatible_vendor_names; *p; p++)
        if(0==strcmp(str, *p))
            return 1;
    return 0;
}

static int check_device_name(const char *str, int *deprecated)
{
    const char * const *p;
    *deprecated = 0;
    for(p = compatible_device_names; *p; p++)
        if(0==strcmp(str, *p))
            return 1;
    *deprecated = 1;
    for(p = deprecated_device_names; *p; p++)
        if(0==strcmp(str, *p))
            return 1;
    return 0;
}

struct unc0xx_scan_item *unc0xx_scan(int unc0xx_id)
{
#ifdef USE_HIDAPI
    struct unc0xx_scan_item *res = NULL;
    struct hid_device_info *devs, *cur_dev;


    devs = hid_enumerate(0, 0);
    for(cur_dev = devs; cur_dev; cur_dev=cur_dev->next) {
        char ss[80]; /* simple string */
        int depr;

        if(cur_dev->vendor_id != UNC_VENDOR_ID) continue;
        if(cur_dev->product_id != UNC_DEVICE_ID) continue;

        simple_from_wchar(cur_dev->manufacturer_string, ss);
        if(!check_vendor_name(ss))
            continue;

        simple_from_wchar(cur_dev->product_string, ss);
        if(!check_device_name(ss, &depr))
            continue;

        /**/
        {
            struct unc0xx_scan_item *tmp;
            tmp = malloc(sizeof(*tmp));
            simple_from_wchar(cur_dev->serial_number, ss);
            extract_sernum(ss, tmp->sernum);
            if(unc0xx_id == 0 || unc0xx_id == tmp->sernum[2]) {
                void *dhdl = (void*)hid_open_path(cur_dev->path);
                if(!dhdl) {
                    free(tmp);
                } else {
                    tmp->next = res;
                    res = tmp;
                    if(depr) {
                        tmp->version = 0;
                        hid_close(dhdl);
                        tmp->dev_hdl = NULL;
                    } else {
                        struct unc0xx_devhdl_usb *hh = create_handler();
                        hh->dev_hdl = dhdl;
                        tmp->dev_hdl = (struct unc0xx_devhdl*) hh;
                        struct unc0xx_global_configuration_str s1;
                        int r = unc0xx_get_configuration(tmp->dev_hdl, &s1);
                        tmp->version = (r == -1) ? 0 : s1.version;
                    }
                }
            } else {
                free(tmp);
            }
        }
    }
    hid_free_enumeration(devs);
    return res;

#else

    struct usb_bus *bus;
    struct usb_device *dev;
    struct unc0xx_scan_item *res = NULL;
    char buf[256];
    int rc;

    for(bus=usb_get_busses(); bus; bus=bus->next) {
        for(dev=bus->devices; dev; dev=dev->next) {
            int matched = 1;
            int deprecated = 0;
            usb_dev_handle *handle;

            if(dev->descriptor.idVendor != UNC_VENDOR_ID) continue;
            if(dev->descriptor.idProduct != UNC_DEVICE_ID) continue;
            
            handle = usb_open(dev);
            if(!handle)
                continue;
            if(!dev->descriptor.iManufacturer) {
                /* no manufacturer string at all */
                matched = 0;
            } else {
                rc = usb_get_string_simple(handle,
                                   dev->descriptor.iManufacturer,
                                   buf, sizeof(buf));
                if(rc <= 0) {
                    matched = 0;
                } else {
                    int nm = check_vendor_name(buf);
                    matched = matched && nm;
                }
            }
            if(!dev->descriptor.iProduct) {
                /* no product string at all */
                matched = 0;
            } else {
                rc = usb_get_string_simple(handle,
                                   dev->descriptor.iProduct,
                                   buf, sizeof(buf));
                if(rc <= 0) {
                    matched = 0;
                } else {
                    int nm = check_device_name(buf, &deprecated);
                    matched = matched && nm;
                }
            }
            if(!dev->descriptor.iSerialNumber) {
                /* no serial number string at all */
                //matched = 0;
                if(matched) deprecated = 1;
            } else {
                rc = usb_get_string_simple(handle,
                                    dev->descriptor.iSerialNumber,
                                    buf, sizeof(buf));
                if(rc <= 0) {
                    //matched = 0;
                    if(matched) deprecated = 1;
                }
            }
            // now it's the Serial in the buf
            if(matched) {
                struct unc0xx_scan_item *tmp;
                tmp = malloc(sizeof(*tmp));
                tmp->dev_hdl = NULL;
                extract_sernum(buf, tmp->sernum);
                if(unc0xx_id == 0 || unc0xx_id == tmp->sernum[2]) {
                    tmp->next = res;
                    res = tmp;
                    if(deprecated) {
                        tmp->version = 0;
                        usb_close(handle);
                    } else {
                        struct unc0xx_devhdl_usb *hh = create_handler();
                        hh->dev_hdl = handle;
                        tmp->dev_hdl = (struct unc0xx_devhdl*) hh;
                        struct unc0xx_global_configuration_str s1;
                        int r = unc0xx_get_configuration(tmp->dev_hdl, &s1);
                        tmp->version = (r == -1) ? 0 : s1.version;
                    }
                } else {
                    usb_close(handle);
                    free(tmp);
                }
            } else { /* not matched, not interesting... */
                usb_close(handle);
            }
        }
    }
    return res;
#endif
}

void unc0xx_destroy_list(struct unc0xx_scan_item *lst)
{
    while(lst) {
        struct unc0xx_scan_item *tmp = lst;
        lst=lst->next;
        if(tmp->dev_hdl)
            unc0xx_close(tmp->dev_hdl);
        free(tmp);
    }
}


/* ------------------- the methods ------------------ */


static int usb_method_close(struct unc0xx_devhdl *h)
{
    struct unc0xx_devhdl_usb *hdl = (struct unc0xx_devhdl_usb *)h;
#ifdef USE_HIDAPI
    hid_close((hid_device*)(hdl->dev_hdl));            
#else
    usb_close(hdl->dev_hdl);
#endif
    destroy_handler(hdl);   
    return 0;
}


static
int usb_method_get(struct unc0xx_devhdl *h, int query_id, void *buf, int len)
{
    struct unc0xx_devhdl_usb *hdl = (struct unc0xx_devhdl_usb *)h;

#ifdef USE_HIDAPI
    ((char*)buf)[0] = query_id;
    return hid_get_feature_report((hid_device*) hdl->dev_hdl, buf, len);
#else
    return
        usb_control_msg(hdl->dev_hdl,
                        USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                        USBRQ_HID_GET_REPORT,
                        USB_HID_REPORT_TYPE_FEATURE << 8 |
                            (query_id & 0xff),
                        0, buf, len, unc0xx_usb_timeout);
#endif
}


static
int usb_method_send(struct unc0xx_devhdl *h, const void *buf, int len)
{
    struct unc0xx_devhdl_usb *hdl = (struct unc0xx_devhdl_usb *)h;

#ifdef USE_HIDAPI
    return hid_send_feature_report((hid_device*) hdl->dev_hdl, buf, len);
#else
    return
        usb_control_msg(hdl->dev_hdl,
                        USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                        USBRQ_HID_SET_REPORT,
                        USB_HID_REPORT_TYPE_FEATURE << 8 |
                            (*(const char*)buf & 0xff),
                        0, (void*)buf, len, unc0xx_usb_timeout);
#endif
}

static int usb_method_takeover_stream(struct unc0xx_devhdl *h)
{

#ifdef USE_HIDAPI
    /* hopefully nothing to be done... well, it doesn't work anyway */
    return 0;
#else
    struct unc0xx_devhdl_usb *hdl = (struct unc0xx_devhdl_usb *)h;

    if(LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP)
    {
        // Detach kernel driver (usbhid) from device interface. (Linux hack)
        usb_detach_kernel_driver_np(hdl->dev_hdl, 0);
        usb_detach_kernel_driver_np(hdl->dev_hdl, 1);
    }
    if(usb_set_configuration(hdl->dev_hdl, 1) < 0){
        unc0xx_set_errormsg(h, usb_strerror(), unc0xx_msg_copy);
        return -1;
    }
    if(usb_claim_interface(hdl->dev_hdl, 0) < 0){
        unc0xx_set_errormsg(h, usb_strerror(), unc0xx_msg_copy);
        return -1;
    }
    return 0;
#endif
}

static int usb_method_read_stream(struct unc0xx_devhdl *h, char *buf)
{
    struct unc0xx_devhdl_usb *hdl = (struct unc0xx_devhdl_usb *)h;

#ifdef USE_HIDAPI
    return hid_read(hdl->dev_hdl, (unsigned char*)buf, 8);
#else
    return usb_interrupt_read(hdl->dev_hdl, 0x81, buf, 8, 0);
#endif
}
