/*
 * UNC0xx host-side software. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc_chown.c
 * This program scans the USB for connected UNC0xx devices and changes
 * the owner/group/mode for the corresponding files.  Unix only.
 *
 * The configuration can be taken from /etc/unc_chown.conf, *OR*
 * from a file given at the commandline, *OR* from the commandline
 * directly.  The configuration is a whitespace-separated list of
 * stanzas, each in the form 'dev_id:uid:gid:mode'.  The dev_id
 * is either a hexadecimal ID for the UNC device (e.g. 7775), or
 * '*' to mean 'all other devices' (this must be the last stanza).
 * Both uid and gid can be given as names as well as numerical values,
 * or can be empty.  The mode is either the octal value such as 0660,
 * or empty.  If no configuration can be found, a compiled-in default
 * is used; originally it is '*::unc:660', but you can give your own
 * default.  There must be either exaclty one command line parameter,
 * or none of them; if the parameter is given, then, if it begins with
 * '/', it is interpreted as the configuraiton file's name; otherwise,
 * it is considered to be the configuration as such.
 *
 * Please note that there's a limit of 32 stanzas; every name can not
 * exceed 31 bytes in length (32 bytes for every string).  You can change
 * these defaults if you like, however.
 *
 * You can run this program from udev, or, if you don't have it,
 * simply by cron.  You can also install the program SUID, it is
 * (hopefully) safe.  Commandline parameters are disabled when the
 * program runs SUID.
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <usb.h>

#if STATIC_VERS == 1
#include "_statpw.h"
#else
#include <pwd.h>
#include <grp.h>
#endif

#include "unc_id.h"

#define MAX_STANZAS 32
#define MAX_NAME_LEN 31
#define DEFAULT_CONFIG_FILE "/etc/unc_chown.conf"
#define DEFAULT_CONFIGURATION "*::unc:660"

#define USB_SYS_ROOT "/dev/bus/usb/"

struct stanza {
    int id;
    int uid;
    int gid;
    int mode;
};
enum fsm_states {
    skip_space, read_id, read_uid, read_gid, read_mode, okay, err
};

static struct stanza configuration[MAX_STANZAS];
static int current = 0;
static char namebuf[MAX_NAME_LEN+1];
static int current_char;
static int line = 1;
static int charpos = 0;
static enum fsm_states the_state = skip_space;
static const char *error_comment = NULL;

static void automata_err_msg(const char *fn)
{
    fprintf(stderr, "%s:%d:%d: %s\n", fn, line, charpos,
                    error_comment ? error_comment : "?!");
}

#if 0
static void print_config()
{
    int i;
    for(i = 0; i < current; i++)
        printf("%4x:%d:%d:%o\n",
               configuration[i].id,
               configuration[i].uid,
               configuration[i].gid,
               configuration[i].mode);
}
#endif

static void complete_id()
{
    char *errptr;
    long res;
    namebuf[current_char] = 0;
    if(*namebuf == '*' && namebuf[1] == 0) {
        configuration[current].id = 0;
        return;
    }
    res = strtol(namebuf, &errptr, 16);
    if(!res || *errptr) {
        error_comment = "Invalid device id";
        the_state = err;
        return;
    }
    configuration[current].id = res;
}

static void complete_uid()
{
    char *errptr;
    long res;
    struct passwd *pw;
    namebuf[current_char] = 0;
    if(!current_char) {
        configuration[current].uid = -1;
        return;
    }
    res = strtol(namebuf, &errptr, 16);
    if(namebuf != errptr && !*errptr) {  /* a number: uid? */
        configuration[current].uid = res;
        return;
    }
    /* okay, not a number */
    pw = getpwnam(namebuf);
    if(!pw) {
        error_comment = "Username not found";
        the_state = err;
        return;
    }
    configuration[current].uid = pw->pw_uid;
}

static void complete_gid()
{
    char *errptr;
    long res;
    struct group *gr;
    namebuf[current_char] = 0;
    if(!current_char) {
        configuration[current].gid = -1;
        return;
    }
    res = strtol(namebuf, &errptr, 16);
    if(namebuf != errptr && !*errptr) {  /* a number: gid? */
        configuration[current].gid = res;
        return;
    }
    /* okay, not a number */
    gr = getgrnam(namebuf);
    if(!gr) {
        error_comment = "Groupname not found";
        the_state = err;
        return;
    }
    configuration[current].gid = gr->gr_gid;
}

static void complete_mode()
{
    char *errptr;
    long res;
    namebuf[current_char] = 0;
    if(!*namebuf) {
        configuration[current].mode = 0;
        current++;
        return;
    }
    res = strtol(namebuf, &errptr, 8);
    if(*errptr || errptr == namebuf || res & 07111) {
        error_comment = "Invalid mode";
        the_state = err;
        return;
    }
    configuration[current].mode = res;
    current++;
}

static void automata_feedchar(int c)
{
    if(the_state == okay || the_state == err)
        return;
    if(c == '\n') {
        line++;
        charpos = 0;
    } else {
        charpos++;
    }
    if(c == 0 || c == -1 /* EOF */) {
        switch(the_state) {
        case read_mode:
            complete_mode();
            the_state = okay;
            break;
        case skip_space:
            the_state = okay;
            break;
        default:
            the_state = err;
        }
        return;
    }
    if(c == ' ' || c == '\n' || c == '\t') {
        if(the_state == read_mode) {
            complete_mode();
            the_state = skip_space;
            return;
        }
        if(the_state != skip_space)
            the_state = err;
        return;
    }
    switch(the_state) {
    case skip_space:
         /* otherwise, go further */
         current_char = 0;
         the_state = read_id;
    case read_id:
         if(c == ':') {
             complete_id();
             current_char = 0;
             the_state = read_uid;
             return;
         }
         break;
    case read_uid:
         if(c == ':') {
             complete_uid();
             current_char = 0;
             the_state = read_gid;
             return;
         }
         break;
    case read_gid:
         if(c == ':') {
             complete_gid();
             current_char = 0;
             the_state = read_mode;
             return;
         }
         break;
    case read_mode:
         if(c == ':') {
             the_state = err;
             return;
         }
         break;
    case okay:
    case err:
         return;
    }
    if(current_char >= sizeof(namebuf) - 1) {
        the_state = err;
        return;         
    }
    namebuf[current_char++] = c;
    namebuf[current_char] = 0;
}

static int config_from_file(FILE* f)
{
    int c;
    while((c = fgetc(f)) != EOF)
        automata_feedchar(c);
    automata_feedchar(-1);
    return the_state == okay;
}

static int config_from_string(const char *s)
{
    while(*s) {
        automata_feedchar(*s);
        s++;
    }
    automata_feedchar(-1);
    return the_state == okay;
}

static const char *compatible_vendor_names[] = {
    UNC_VENDOR_NAME,
    "UNIControllers http://www.unicontrollers.com",
    NULL
};

static const char *compatible_device_names[] = {
    UNC_DEVICE_NAME,
    UNC_DEVICE_NAME_0XX_LEGACY,
    UNC_DEVICE_NAME_001_LEGACY,
    NULL
};

static int check_vendor_name(const char *str)
{
    const char * const *p;
    for(p = compatible_vendor_names; *p; p++)
        if(0==strcmp(str, *p))
            return 1;
    return 0;
}

static int check_device_name(const char *str)
{
    const char * const *p;
    for(p = compatible_device_names; *p; p++)
        if(0==strcmp(str, *p))
            return 1;
    return 0;
}

static int hexdig2num(char c)
{
    if(c >= '0' && c <= '9') return c - '0';    
    if(c >= 'A' && c <= 'F') return c - 'A' + 10;    
    if(c >= 'a' && c <= 'f') return c - 'a' + 10;    
    return -1;
}

static int extract_sernum(const char *str, int sernum[3])
{
    int p = 0;
    int n = 0;
    sernum[0] = sernum[1] = sernum[2] = -1;
    for(; *str && p<3; str++) {
        if(*str == ':') {
            sernum[p] = n;
            p++;
            n = 0;
        } else {
            int d = hexdig2num(*str);
            if(d == -1) return 0;
            n = n*16 + d;
        }
    }
    if(!*str) {
        sernum[p] = n;
        p++;
    }
    return p==3;
}

static void handle_device(int id, const char *dir, const char *fname)
{
    int conf = -1;
    int i;
    char *path;
    for(i = 0; i < current; i++)
        if(configuration[i].id == id || configuration[i].id == 0) {
            conf = i;
            break;
        }

    if(conf == -1)
        return;

    path = malloc(sizeof(USB_SYS_ROOT) + strlen(dir) + strlen(fname) + 3);
    *path = 0;
    strcat(path, USB_SYS_ROOT);
    strcat(path, "/");
    strcat(path, dir);
    strcat(path, "/");
    strcat(path, fname);

    /* fprintf(stderr, "%04x: %s\n", id, path); */

    if(configuration[conf].uid != -1 || configuration[conf].gid != -1) {
        int uid = configuration[conf].uid;
        int gid = configuration[conf].gid;
        int r = chown(path, uid, gid);
        if(r == -1) {
            perror(path);
            /* we don't stop here, we'll try to do chmod as well */
        }
    }
    if(configuration[conf].mode) {
        int r = chmod(path, configuration[conf].mode);
        if(r == -1) {
            perror(path);
        }
    }
    free(path);
}

static void scan()
{
    struct usb_bus      *bus;
    struct usb_device   *dev;
    char buf[256];
    int sernum[3];
    int rc;

    for(bus=usb_get_busses(); bus; bus=bus->next) {
        for(dev=bus->devices; dev; dev=dev->next) {
#if 0
            fprintf(stderr, "Device: %04x / %04x file: %s/%s\n",
                   dev->descriptor.idVendor,
                   dev->descriptor.idProduct,
                   bus->dirname, dev->filename);
#endif
            usb_dev_handle *handle;

            if(dev->descriptor.idVendor != UNC_VENDOR_ID) continue;
            if(dev->descriptor.idProduct != UNC_DEVICE_ID) continue;

            handle = usb_open(dev);
            if(!handle)
                continue;

            if(!dev->descriptor.iManufacturer ||
                !dev->descriptor.iProduct ||
                !dev->descriptor.iSerialNumber)
            {
                goto not_a_match;
            }
            
            rc = usb_get_string_simple(handle,
                               dev->descriptor.iManufacturer,
                               buf, sizeof(buf));
            if(rc <= 0 || !check_vendor_name(buf))
                goto not_a_match;

            rc = usb_get_string_simple(handle,
                               dev->descriptor.iProduct,
                               buf, sizeof(buf));
            if(rc <= 0 || !check_device_name(buf))
                goto not_a_match;

            rc = usb_get_string_simple(handle,
                                dev->descriptor.iSerialNumber,
                                buf, sizeof(buf));
            if(rc <= 0 || !extract_sernum(buf, sernum)) 
                goto not_a_match;

            handle_device(sernum[2], bus->dirname, dev->filename);
        not_a_match:
            usb_close(handle);
        }
    }
}

int main(int argc, const char * const * argv)
{

    const char *conf_file = NULL;
    const char *conf_string = NULL;

    if(getuid() == geteuid() && getgid() == getegid()) {
        /* okay, we're not suid so we can analyse the commandline */
        if(argc > 2) {
            fprintf(stderr, "Too many parameters\n");
            return 1;
        }
        if(argc == 2) {
            if(argv[1][0] == '/')
                conf_file = argv[1];
            else
                conf_string = argv[1];
        } else {
            conf_file = DEFAULT_CONFIG_FILE;
            conf_string = DEFAULT_CONFIGURATION;
        }
    } else {
        /* we're running either SUID or SGID (or both);
         * hence we must refuse any commandline */
        if(argc > 1) {
            fprintf(stderr,
                "No commandline params are allowed in SUID/SGID mode\n");
            return 1;
        }
        conf_file = DEFAULT_CONFIG_FILE;
        conf_string = DEFAULT_CONFIGURATION;
    }

    if(conf_file) {
        FILE* f = fopen(conf_file, "r");
        if(f) {
            conf_string = NULL; /* no conf. string, we're using the file */
            if(!config_from_file(f)) {
                automata_err_msg(conf_file);
                return 1;
            }
            fclose(f);
        } else {
            if(!conf_string) {
                perror(conf_file);
                return 1;
            }
        }
    }
    if(conf_string) {
        if(!config_from_string(conf_string)) {
            automata_err_msg("conf");
            return 1;
        }
    }

    usb_init();
    usb_find_busses();
    usb_find_devices();
    scan();

    return 0;
}
