/*
 * UNC0xx firmware. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file control.c
 * Control request handling and answering
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <avr/eeprom.h>

#include "unc_prot.h"
#include "unc_prim.h"
#include "unc_1w.h"
#include "misc.h"
#include "feat_set.h"
/*#include "twowire.h"*/

#include "control.h"


/* The following variables store the status of the current data transfer */
/*static uint8_t current_address;*/
static uint8_t request_buffer[64];  /* we'll never need more */
static uint8_t bytes_remaining;
static uint8_t eeprom_portion = 0;
static uint8_t extra_state_page = 0;
uint8_t *transfer_data_pointer;


static void finalize_incoming_request()
{
    switch(request_buffer[1]) {
        case uncreqcmd_set_hwconfig:
            eeprom_write_block(request_buffer+4, (uint8_t *)0+16, 4); 
            /* intentionally no break here */
        case uncreqcmd_restore_from_eeprom:
            init_primary_hardware();
            /*init_timer();*/
            break;
        case uncreqcmd_set_userconfig:
            memcpy_4b(global_configuration.user_config_bytes,
                      request_buffer+4);
            break;
        case uncreqcmd_set_id: /* set device id */
            eeprom_write_block(request_buffer+4, (uint8_t *)0+4, 2 /*was 4*/); 
            break;
        case uncreqcmd_set_reactions: /* set reaction register */
            memcpy_4b(
                global_configuration.react_bytes[request_buffer[2] & 0x07],
                request_buffer+4);
            break;
        case uncreqcmd_save_to_eeprom: /* save to eeprom */
            save_registers_to_eeprom();
            break;
        case uncreqcmd_restore_factory:
            restore_factory_defaults();
            break;
        case uncreqcmd_action: /* perform action */
            perform_action(request_buffer[5], request_buffer[4]);
            break;
        case uncreqlcmd_delayed_action:
            //schedule_delayed_action(request_buffer + 4);
            schedule_delayed_action_cycles(request_buffer[4],
                                           request_buffer[3],
                                           request_buffer[2],
                                           *(uint32_t*)(request_buffer + 8));
            break;
        case uncreqcmd_clear_msg:
            /* global_state.buf_start = global_state.buf_end = 0; */
            global_state_drop_message();
            break;
        case uncreqlcmd_write_e_reg:
            if(request_buffer[2] < 48) {
                eeprom_write_block(request_buffer+3,
                    (uint8_t *)128 + request_buffer[2] * 8, 8); 
            }
            break;
        case uncreqcmd_reset_1wire:
            unc1w_init();
            break;
    }
}

void start_incoming_request(uint8_t cmdcode)
{
    /* we only use two requests: 8 and 16 bytes */
    bytes_remaining = (cmdcode == uncreq_long_command) ? 16 : 8;
    /*current_address = 0;*/
    transfer_data_pointer = request_buffer;
}

uint8_t process_incoming_data(uint8_t *data, uint8_t len)
{
    uint8_t i;

    if(bytes_remaining < len)
        len = bytes_remaining;

    for(i = 0; i<len; i++)
        *transfer_data_pointer++ = data[i];

    bytes_remaining -= len;

    if(bytes_remaining == 0) {
        finalize_incoming_request();
        return 1;
    }
    return 0;
}


static void start_extra_state_request()
{
    if(extra_state_page == 0) {
        extra_state.page_seq = 0x80;   /* anothre page's to come */
        extra_state.page_version = 1;  /* it's the _1 type */
        transfer_data_pointer = (uint8_t*)(&extra_state);
        extra_state_page = 1;
    } else {
        delay_state.page_seq = 0x01;   /* the #1 is the last */
        delay_state.page_version = 2;  /* it's the _2 type */
        transfer_data_pointer = (uint8_t*)(&delay_state);
        extra_state_page = 0;
    }
}

uint8_t start_data_request(uint8_t cmdcode)
{
    switch((request_buffer[0] = cmdcode)) {
        case uncreq_query_fullstate:
            extra_state_page = 0; // it's a useful side effect
            sync_global_state();
            transfer_data_pointer = (uint8_t*)(&global_state);
            return sizeof(global_state);
        case uncreq_query_configuration:
            eeprom_portion = 0; // it's a useful side effect
            global_configuration.version = UNC_VERSION;
            transfer_data_pointer = (uint8_t*)(&global_configuration);
            return sizeof(global_configuration);
        case uncreq_zero_query_counters:
            transfer_data_pointer = request_buffer;
            zero_counters((uint16_t*)(request_buffer+1));
            return sizeof(request_buffer);  /* 1 byte + 8x2 b. counters */
        case uncreq_query_extrastate:
            /* sync_global_state(); // no need for this */
            start_extra_state_request();
            return sizeof(extra_state);
        case uncreq_query_eeprom:
        default: /* to make the stupid compiler happy */
#if UNC_FEATURE_DUMP_EEPROM 
            eeprom_read_block(request_buffer,
                (uint8_t*)((uint16_t)64 * (uint16_t)eeprom_portion), 64);
            ++eeprom_portion;
            eeprom_portion &= 7;
            transfer_data_pointer = request_buffer;
            return 64;
#else
            return 0;
#endif
    }
}

