/*
 * UNC0xx firmware. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc_1w.c
 * Implemetation of the 1-Wire(R) support
 *
 * Written by Andrey V. Stolyarov
 *
 */

//#include <avr/wdt.h>
#include <avr/eeprom.h>

#include "feat_set.h"

#include "print.h"
#include "unc_prim.h"

#include "crc8.h"
#include "onewire.h"


#include "unc_1w.h"



#define ONEW_DEV_COUNT 6

#define MINIMAL_DROP_DELAY 48


/* DS18x20-related defines */
#define DS18S20_ID           0x10
#define DS18B20_ID           0x28
#define DS18X20_CONVERT_T    0x44
#define DS18X20_READ         0xBE

/* DS2413-related defines */
#define DS2413_ID            0x3a
#define DS2413_PIO_READ      0xF5
#define DS2413_PIO_WRITE     0x5A

static uint8_t id_tmarks[ONEW_DEV_COUNT] = { 0, 0, 0, 0, 0, 0 };

static void print_id(uint8_t charcode, uint8_t checkmask, uint8_t *id)
{
    if(global_configuration.user_config_bytes[2] & checkmask) {
        int8_t j;
        print_char('W');
        print_char(charcode);
        for(j=7; j>=0; j--)
            print_hex_byte(id[j]);
        print_char('\n');
    }
}

static uint8_t e_reg_matches_id(uint8_t reg, const uint8_t *id)
{
    uint8_t k;
    for(k=0; k<7; k++) {
        uint8_t b = eeprom_read_byte((uint8_t*) 128 + reg*8 + k);
        if(b != id[k]) return 0;
    }
    return 1;
}

static void handle_connected_ibutton(const uint8_t *id)
{
    int8_t i;
#if UNC_FEATURE_HANDLE_CONN_BIT  /* check the conf. bit 'handle connected' */
    if(!(global_configuration.user_config_bytes[2] & 0x08))
        return;
#endif
    for(i=0; i<48; i++) {
        uint8_t b = eeprom_read_byte((uint8_t*) 128 + i*8 + 7);
        if(b <= 0x2f) {  /* iButton id stored */
            if(e_reg_matches_id(i, id)) {
                perform_action(0x0e, b);
            }
        } /* else intentionally removed to save bytes */
        if(b & 0x80) {
            /* stored ids must occupy a continuous area */
            /* starting with the E00 register */
            /* if this register doesn't comply, no further will */
            return;
        }
    }
}

struct temp_slave_reg {
    int16_t hi_wm, lo_wm;
    uint8_t lsb_a_hi_wm;
    uint8_t lsb_a_lo_wm;
    uint8_t msb_a;
    uint8_t ff;
};

static void handle_temp_change(int16_t last, int16_t next, uint8_t *id)
{
    uint8_t last_invalid =
        (uint16_t)(last) == (uint16_t)0x8000 || last >= 0x0540;
    uint8_t k;
#if UNC_FEATURE_TEMP_ACTION_SCAN_BIT 
    if(!(global_configuration.user_config_bytes[2] & 0x04))
        return;
#endif
    for(k=0; k<47; k++) {
        uint8_t b = eeprom_read_byte((uint8_t*) 128 + k*8 + 7);
        if(b == 0x70 && e_reg_matches_id(k, id)) {
            do {
                struct temp_slave_reg s;
                eeprom_read_block(&s, (uint8_t*)128+(k+1)*8, 8);
                if(s.ff != 0x7e)
                    break;
                if((last_invalid || last > s.lo_wm) && next <= s.lo_wm)
                    perform_action(s.msb_a, s.lsb_a_lo_wm);
                else
                if((last_invalid || last < s.hi_wm) && next >= s.hi_wm)
                    perform_action(s.msb_a, s.lsb_a_hi_wm);
            } while(++k < 47); /* note we query the (k+1)'th reg */
            break;
        } else
        if(b & 0x80) {
            break;
        }
    }
}

static void register_id(uint8_t id[8])
{
    int8_t free_pos = -1, found_pos = -1;
    int8_t i, j;
    /* for(i=ONEW_DEV_COUNT-1; i>=0; i--) { // worse! */
    /* for(i=ONEW_DEV_COUNT; i>0; ) { i--; // worse! */
    for(i=0; i<ONEW_DEV_COUNT; i++) {
        if(extra_state.ow_ids[i][0] == 0) {
            /* empty slot */
            free_pos = i;
        } else {
            for(j=0; j<8; j++)
                if(id[j] != extra_state.ow_ids[i][j])
                    goto no_match;
            /* here is it! */
            found_pos = i;
            break;
        no_match:
            ;
        }
    } 
    if(found_pos == -1) { /* not found => new connected device */
        if(free_pos == -1) /* no free slots -> ignoring silently */
            return;
        print_id('+', 0x40, id);
        for(j=0; j<8; j++)
            extra_state.ow_ids[free_pos][j] = id[j];
        found_pos = free_pos;
#if 0    /* no actual reason for the check, and gains 6 bytes */
        if(id[0] == 0x01) { /* iButton "key" */
            handle_connected_ibutton(id);
        }
#else
        handle_connected_ibutton(id);
#endif
    }
    id_tmarks[found_pos] = global_state.timemark_bytes[0];
}

static void do_read_temperature(uint8_t i)
{
    uint8_t buf[9];
    uint8_t j;
    ow_command(DS18X20_READ, extra_state.ow_ids[i]);
    for(j=0; j<9; j++)
        buf[j] = ow_byte_rd();
    if(crc8(buf, 8) == buf[8]) {
        int16_t t = buf[1];
        t <<= 8;
        t |= buf[0];
        if(extra_state.ow_ids[i][0] == DS18S20_ID)
            t <<= 3;
        handle_temp_change(
            extra_state.ow_temperature[i], t,
            extra_state.ow_ids[i]);
        extra_state.ow_temperature[i] = t;
    }
}

static void do_read_keystate(uint8_t i)
{
    int8_t j = 7;
    uint8_t b1, b2;
    do {
        ow_command(DS2413_PIO_READ, extra_state.ow_ids[i]);
        b1 = ow_byte_rd();
        j--;
    } while(b1 == 0xff && j);
    for(; j; j--) {
        b2 = ow_byte_rd();
        if(b1 == b2 && (b1 & 0x0f) == ((~b1 >> 4) & 0x0f)) {
            extra_state.ow_data_bytes[i][0] = b1;
            extra_state.ow_data_bytes[i][1] = 0;
            break;
        }
        if(b1 == 0xff) {
            extra_state.ow_data_words[i] = 0xffff;
            break;
        }
    }
    ow_reset();
}

void unc1w_init()
{
    uint8_t k;
    //ow_set_bus(&PINB, &PORTB, &DDRB, PB0);
    for(k = 0; k < ONEW_DEV_COUNT; k++) {
        extra_state.ow_ids[k][0] = 0;
        /* extra_state.ow_temperature[k] = 0x8000; */
    }
}

void unc1w_hook()
{
    static uint8_t counter = 0;
    uint8_t id[8], diff, i, dropdelay;

#if UNC_FEATURE_READ_1WIRE_DATA_BIT 
    uint8_t scanex;
#endif

    if(!(global_configuration.user_config_bytes[2] & 0x80))
        return;

    ++counter;
    counter &= 0x0f;

    if(!counter) { /* once in 16 ticks */
        if(ow_reset() == 0) {
            for(diff = OW_SEARCH_FIRST; diff != OW_LAST_DEVICE;) {
                diff = ow_rom_search(diff, id);
                if(diff == OW_PRESENCE_ERR || diff == OW_DATA_ERR)
                    return; /* don't perform anything else on this tick */
                register_id(id);
            }
        }
        /* drop_absent_ids(); */
    }
    /* scan_extra_info(); */
    /* now we need to scan for extra info, if any, and drop the absent */

#if UNC_FEATURE_READ_1WIRE_DATA_BIT 
    scanex = global_configuration.user_config_bytes[2] & 0x10; 
#endif
    dropdelay = (global_configuration.user_config_bytes[1] & 0xf0) | 0x0f; 
    if(dropdelay < MINIMAL_DROP_DELAY)
        dropdelay = MINIMAL_DROP_DELAY;

    for(i = ONEW_DEV_COUNT; i > 0; ) { i--;

        if(!extra_state.ow_ids[i][0])
            continue;

        diff = (uint8_t)(global_state.timemark_bytes[0] - id_tmarks[i]);
        if(diff > dropdelay) {
            print_id('-', 0x20, extra_state.ow_ids[i]);
            extra_state.ow_ids[i][0] = 0;
            continue;
        }

#if UNC_FEATURE_READ_1WIRE_DATA_BIT 
        if(scanex && diff < 13)
#else
        if(diff < 13)
#endif
            switch(extra_state.ow_ids[i][0]) { 
                case DS18S20_ID:
                case DS18B20_ID:
                    if(counter == 3)  /* request measurement */
                        ow_command(DS18X20_CONVERT_T, extra_state.ow_ids[i]);
                    else
                    if(counter == 11)
                        do_read_temperature(i);
                    break;
                case DS2413_ID:
                    if(counter == 7)
                        do_read_keystate(i);
                    break;
#if 0   /* no real need for this, it will be ignored anyway...
                                 and we gain 22 bytes here! */
                default:
                    extra_state.ow_data_words[i] = 0x8000;
#endif
            }
    }
}

#if 0
void replay_70_reg(uint8_t reg_n)
{
    uint8_t k;
    for(k = 0; k < 4; k++) { /* XXXX */
        if(e_reg_matches_id(reg_n, extra_state.ow_ids[k])) {
            int16_t temp = extra_state.ow_temperature[k];
            if((uint16_t)(temp) == (uint16_t)0x8000 || temp >= 0x0540)
                return;
            handle_temp_change(0x8000, temp, extra_state.ow_ids[k]);
            return;
        }
    } 
}
#endif

void perform_1wire_action(uint8_t action)
{
    uint8_t regbuf[8];
    uint8_t ereg = action >> 2;
    /*action &= 0x03;*/
    action |= ~0x03;
    uint8_t iters = 0;
    uint8_t statbyte;
    if(ereg > 47)
        return;
    eeprom_read_block(regbuf, (uint8_t*)128+ereg*8, 8);
#if 0     /* Why the hell should we care? Nothing will explode anyway! */
    if(regbuf[7] != 0x60)
        return;
#endif
    regbuf[7] = crc8(regbuf, 7);  /* restore the CRC byte */
    do {
        ow_command(DS2413_PIO_WRITE, regbuf);
        ow_byte_wr(action);
        ow_byte_wr(~action);
        statbyte = ow_byte_rd();
        ow_byte_rd();
        ow_reset();
    } while(--iters && statbyte != 0xAA);
}
