/* 
 * UNC0xx firmware. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file usb_io.c
 * Implements the usb io functions called by the usbdrv
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"

#include "unc_prot.h"
#include "unc_prim.h"
#include "control.h"

#include "usb_io.h"


/*#include "usbdrv.c"*/


PROGMEM const char usbHidReportDescriptor[78] = {
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)

    0x85, uncreq_query_eeprom,     //   REPORT_ID (uncreq_query_eeprom)
    0x95, 0x3f,                    //   REPORT_COUNT (64)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x80, 0x01,              //   FEATURE (Data,Ary,Abs,Vol,Buf)

    0x85, uncreq_command,          //   REPORT_ID (uncreq_command)
    0x95, 0x07,                    //   REPORT_COUNT (7)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x80, 0x01,              //   FEATURE (Data,Ary,Abs,Vol,Buf)

    0x85, uncreq_long_command,     //   REPORT_ID (uncreq_long_command)
                                   // NB: the same as uncreq_com_poll
    0x95, 0x0f,                    //   REPORT_COUNT (15)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x80, 0x01,              //   FEATURE (Data,Ary,Abs,Vol,Buf)

    0x85, uncreq_zero_query_counters,// REPORT_ID (uncreq_zero_query_counters)
    0x95, 0x10,                    //   REPORT_COUNT (16)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0xa0, 0x01,              //   FEATURE (Data,Ary,Abs,NPrf,Vol,Buf)

    0x85, uncreq_query_fullstate,  //   REPORT_ID (uncreq_query_fullstate)
    0x95, 0x3f,                    //   REPORT_COUNT (63)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x00, 0x01,              //   FEATURE (Data,Ary,Abs,Buf)

    0x85, uncreq_query_configuration,// REPORT_ID (uncreq_query_configuration)
    0x95, 0x3f,                    //   REPORT_COUNT (63)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x00, 0x01,              //   FEATURE (Data,Ary,Abs,Buf)

    0x85, uncreq_query_extrastate, //   REPORT_ID (uncreq_query_extrastate)
    0x95, 0x3f,                    //   REPORT_COUNT (63)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x00, 0x01,              //   FEATURE (Data,Ary,Abs,Buf)

#if 0   /* we don't know why this doesn't work, but, being uncommented,
           this causes Windows not to recognize our device.  If you know
           how to properly declare the 'interrupt input' stream here,
           please let us know!
         */
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x81, 0x06,                    //   INPUT (Data,Var,Rel)
#endif
    0xc0                           // END_COLLECTION
};



/* usbFunctionWrite() is called when the host sends data */
uint8_t usbFunctionWrite(uint8_t *data, uint8_t len)
{
    return process_incoming_data(data, len);
}

/* usbFunctionSetup() called when a new request arrives */
/* Functions from the control.[ch] module return ((uint8_t)-1) as
   indication for the caller to use functions to transfer data.
   This -1 is transferred unchanged to the usbdrv, instead of
   using USB_NO_MSG constant, as they are the same.  Should this 
   change one day, either using something else instead of -1, or
   usbMsgLen becoming other than uint8_t, we'll have to add another
   if to the usbFunctionSetup() code.
 */
usbMsgLen_t usbFunctionSetup(uint8_t data[8])
{
    usbRequest_t *rq = (void *)data;

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {
                                               /* HID class request */
        // request_buffer[0] = rq->wValue.bytes[0];  // this is 'cmdcode'
        if(rq->bRequest == USBRQ_HID_GET_REPORT) {
            uint8_t len = start_data_request(rq->wValue.bytes[0]);
            if(len < 1) {
                return USB_NO_MSG;
            } else {
                usbMsgPtr = transfer_data_pointer;
                return len;
            }
        }else if(rq->bRequest == USBRQ_HID_SET_REPORT){
            start_incoming_request(rq->wValue.bytes[0]);
            return USB_NO_MSG;  /* use usbFunctionRead()/..Write() */
        }
    }else{
        /* ignore vendor type requests, we don't use any */
    }
    return 0;
}


void process_usb_interrupt_input()
{
    uint8_t c;
    if(!(global_configuration.user_config_bytes[1] & 0x01))
        return;
    if(global_state.buf_end == global_state.buf_start)
        return;
    if(!usbInterruptIsReady())
        return;
    c = (global_state.buf_end > global_state.buf_start) ?
        (global_state.buf_end - global_state.buf_start) :
        (32 - global_state.buf_start);
    if(c > 8) c = 8;
    usbSetInterrupt(global_state.message_buffer + global_state.buf_start, c);
    global_state.buf_start += c;
    global_state.buf_start &= 0x1f;
}
