/*
 * UNC0xx GUI-based software. (c) Unicontrollers, Ltd., 2012-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file devopdlg.cpp
 * Implementation of the "Device options" dialog
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <wx/dialog.h>
#include <wx/msgdlg.h>
#include <wx/gbsizer.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/choice.h>


#include "config.hpp"
#include "glue.hpp"
#include "devopdlg.hpp"


struct DlgListItem {
    int id;
    class DevOptionsDialog *dlg;
    DlgListItem *next;
};
static DlgListItem *first = 0;

class DevOptionsDialog: public wxDialog {
    int id;
    UncCtrlsetCollection *the_controls;
    DeviceConfig *conf;
    wxGridBagSizer *grid;
    wxTextCtrl *devname;
    wxChoice *modules[2];
    int last_mod_data[2];
    wxStaticText *labels[12];
    wxTextCtrl *inputs[12];
public:
    DevOptionsDialog(int id, UncCtrlsetCollection *controls);
    ~DevOptionsDialog();

private:
    static wxString MakeCaption(int id);
    void Sync();
    void SyncLineLabels();
    void OnOk(wxCommandEvent& event);
    void OnCancel(wxCommandEvent& event);
    void OnChoice(wxCommandEvent& event);

    DECLARE_EVENT_TABLE();
};

DevOptionsDialog::DevOptionsDialog(int a_id, UncCtrlsetCollection *controls)
    : wxDialog(NULL, -1, MakeCaption(a_id), wxDefaultPosition,
               wxSize(600,500),
               wxCAPTION|wxRESIZE_BORDER|wxMINIMIZE_BOX|wxDIALOG_NO_PARENT),
      id(a_id), the_controls(controls), conf(GetDeviceConfig(a_id)), grid(0)
{
    grid = new wxGridBagSizer(2, 2);
    grid->SetRows(15);
    grid->SetCols(3);
    //grid->AddGrowableCol(0);
    grid->AddGrowableCol(2);

    wxString slot_choices[] = { wxT("None"), wxT("Input"), wxT("Output") };
    wxBoxSizer *box = new wxBoxSizer(wxHORIZONTAL);
    box->Add(new wxStaticText(this, wxID_STATIC, wxT("Name: ")));
    devname = new wxTextCtrl(this, wxID_NONE, wxT(""),
                             wxDefaultPosition, wxSize(100, -1));
    box->Add(devname);
    box->Add(new wxStaticText(this, wxID_STATIC, wxT("Slot 1: ")));
    modules[0] = new wxChoice(this, wxID_ANY,
                 wxDefaultPosition, wxDefaultSize, 3, slot_choices);
    box->Add(modules[0]);
    box->Add(new wxStaticText(this, wxID_STATIC, wxT("Slot 2: ")));
    modules[1] = new wxChoice(this, wxID_ANY,
                 wxDefaultPosition, wxDefaultSize, 3, slot_choices);
    box->Add(modules[1]);
    grid->Add(box, wxGBPosition(0, 0), wxGBSpan(1, 3), wxGROW); 

    for(int i = 0; i<12; i++) {
        labels[i] = new wxStaticText(this, wxID_STATIC, wxT(""));
        grid->Add(labels[i], wxGBPosition(i+1, 0));
        inputs[i] = new wxTextCtrl(this, wxID_NONE, wxT(""),
                                   wxDefaultPosition, wxSize(300, -1));
        grid->Add(inputs[i], wxGBPosition(i+1, 2), wxDefaultSpan, wxGROW);
    }
    Sync();
    wxSizer *s = CreateSeparatedButtonSizer(wxOK|wxCANCEL);
    grid->Add(s, wxGBPosition(13, 0), wxGBSpan(1, 3), wxGROW);

    grid->SetSizeHints(this);
    SetSizer(grid);
    Fit();

    DlgListItem *tmp = new DlgListItem;
    tmp->next = first;
    tmp->id = a_id;
    tmp->dlg = this;
    first = tmp;
}

DevOptionsDialog::~DevOptionsDialog()
{
    DlgListItem **p = &first;
    while(*p) {
        if((*p)->dlg == this) {
            DlgListItem *tmp = *p;
            *p = tmp->next;
            delete tmp;
            break; // definitely we can't be there two times :-)
        } else {
            p = &((*p)->next);
        }
    }
}

wxString DevOptionsDialog::MakeCaption(int id)
{
    wxString s;
    s.Printf(wxT("Options for: %04x"), id);
    return s;
}

void DevOptionsDialog::Sync()
{
    int i;
    devname->ChangeValue(wxString(conf->GetName(), wxConvLocal));
    for(i=0; i<2; i++) {
        switch(conf->GetModuleType(i)) {
            case lnt_none:   last_mod_data[i] = 0; break;
            case lnt_input:  last_mod_data[i] = 1; break;
            case lnt_output: last_mod_data[i] = 2; break;
        }
        modules[i]->SetSelection(last_mod_data[i]);
    }
    SyncLineLabels();
    for(i=0; i<12; i++) {
        inputs[i]->ChangeValue(wxString(conf->GetLineName(i), wxConvLocal));
    }
}

void DevOptionsDialog::SyncLineLabels()
{
    int i;
    for(i=0; i<12; i++) {
        wxString s;
        int lt;
        lt = (i < 4) ?  2 : modules[i/4 - 1]->GetSelection();
        switch(lt) {
        case 0:
            s = wxT("(no module)");
            break;
        case 1:
            s.Printf(wxT("Input #%d"), i-4); 
            break;
        case 2:
            s.Printf(wxT("Output #%d"), i); 
            break;
        }
        labels[i]->SetLabel(s);
    }
}

void DevOptionsDialog::OnOk(wxCommandEvent& event)
{
    int i;
    bool modified = false;
    static const line_type lnt[3] = { lnt_none, lnt_input, lnt_output };
    if(devname->IsModified()) {
        conf->SetName(devname->GetValue().mb_str(wxConvLocal));
        modified = true;
    }
    for(i=0; i<2; i++) {
        if(modules[i]->GetSelection() != last_mod_data[i]) {
            conf->SetModuleType(i, lnt[modules[i]->GetSelection()], false);
            modified = true;
        }
    }
    for(i=0; i<12; i++) {
        if(inputs[i]->IsModified()) {
            conf->SetLineName(i, inputs[i]->GetValue().mb_str(wxConvLocal));
            modified = true;
        }
    }
    if(modified) {
        conf->Save();
        the_controls->Reconfigure(id);
        the_controls->CallRescan();
    }
    delete this;
}

void DevOptionsDialog::OnCancel(wxCommandEvent& event)
{
    delete this;
}

void DevOptionsDialog::OnChoice(wxCommandEvent& event)
{
    SyncLineLabels();
}


BEGIN_EVENT_TABLE(DevOptionsDialog, wxDialog)
    EVT_BUTTON(wxID_OK,     DevOptionsDialog::OnOk)
    EVT_BUTTON(wxID_CANCEL, DevOptionsDialog::OnCancel)
    EVT_CHOICE(wxID_ANY,    DevOptionsDialog::OnChoice)
END_EVENT_TABLE()



///////////////////////////////////////////////////////////////////////

void LaunchDeviceOptionsDialog(int id, class UncCtrlsetCollection *controls)
{
#if 0
    wxString msg;
    msg.Printf(wxT("Device: %04x"), id);
    wxMessageBox(msg, wxT("Just a message"), wxOK|wxCENTRE, parent);
#endif
    for(DlgListItem *p = first; p; p = p->next) {
        if(p->id == id) {
            p->dlg->Show();
            p->dlg->Raise();
            return;
        }
    }
    DevOptionsDialog *dlg = new DevOptionsDialog(id, controls);
    dlg->Show();
}
